local L = LibStub("AceLocale-3.0"):GetLocale("JeypLoathebHealSorter")

JeypLoathebHealSorter.commPrefix = "JeypLoathebHS";
JeypLoathebHealSorter.sPlayerName = UnitName("player")
JeypLoathebHealSorter.sPlayerGUID = UnitGUID("player")
JeypLoathebHealSorter.sActiveSpellName = nil
JeypLoathebHealSorter.sCooldownSpellName = nil
JeypLoathebHealSorter.sSporeSpellName = nil
JeypLoathebHealSorter.sAddonVersionID = "0.1.7"
JeypLoathebHealSorter.bPrintDebugInfo = false
JeypLoathebHealSorter.bPrintLog = false
JeypLoathebHealSorter.bProfiling = false

local tLoathebSpells = {
	sActiveSpellName = L["Corrupted Mind"], -- Limits Healing casts to 1 / min
	sCooldownSpellName = L["Corrupted Mind"], -- 1 min cooldown
	sSporeSpellName = L["Fungal Bloom"], -- Crit Buff
}
local tTestSpells = {
	sActiveSpellName = L["Power Word: Fortitude"],
	sCooldownSpellName = L["Renew"],
	sSporeSpellName = L["Power Word: Shield"],
}

-- ACE CALLBACKS
function JeypLoathebHealSorter:OnInitialize()
	self:InitOptions()
	self:InitHealerFrames()
	
	self.IconLauncher = LibStub("LibDataBroker-1.1"):NewDataObject("JeypLoathebHealSorter", {
		type = "launcher",
		text = "JeypLoathebHealSorter",
		icon = "Interface/Icons/spell_shadow_auraofdarkness",
		OnClick = function(self, button)
			JeypLoathebHealSorter:IconClick()
		end,
		OnEnter = function(self)
			GameTooltip:SetOwner(self, "ANCHOR_LEFT")
			JeypLoathebHealSorter:DrawTooltip()
		end,
		OnLeave = function(self)
			GameTooltip:Hide()
		end
	})
	
	self.icon = LibStub("LibDBIcon-1.0")
	self.icon:Register("JeypLoathebHealSorter", self.IconLauncher, self.db.global)
	
	self.cRosterUpdateTimer = nil
	self.iGreatestTimestamp = 0
	self.iGreatestTimestampSeparator = 0
	self.sRaidSeparator = ","
	self.fLastTimeInCombat = GetServerTime()
	self.bReadyForReset = false
	self.bRunningTests = false
	self.bSetLoadedInFight = false
	self.tHealerOrder = {}
	self.tNonHealerList = {}
	self.tNonHealerOrder = {}
	self.bIsInNaxx = false
	
	self:RegisterComm(self.commPrefix);
	SLASH_JEYPLOATHEBHS1 = "/jlhs"
	SLASH_JEYPLOATHEBHS2 = "/jeyploathebhealsorter"
	SlashCmdList["JEYPLOATHEBHS"] = function(msg)
		msg = string.lower(msg)
		msg = { string.split(" ", msg) } or msg
		
		if msg[1] == "export" then 
			--self:ExportDebug() Disabled - easier to just load WTF file
		elseif msg[1] == "test" then
			self:RunTestCase(msg[2])
		else
			self:IconClick()
		end
	end 
	
	self:SetSpellNames()
	self:UpdateFontSize()
	self:HideAnchor()
	
	--Clear any log older than 2 weeks
	self:ClearOldLogs()
	
	self:AddToLogInit("sAddonVersionID", self.sAddonVersionID)
	self:AddToLogInit("bTestMode", self.db.global.bTestMode)
	self:AddToLogInit("bSendData", self.db.global.bSendData)
	self:AddToLogInit("sSeparator", self.db.global.sSeparator)
	self:AddToLogInit("sHealerList", self.db.global.sHealerList)
end

function JeypLoathebHealSorter:RunTestCase(iCase)
	if iCase == nil then iCase = 1 end
	iCase = tonumber(iCase)
	if iCase == nil then iCase = 1 end
	print("Starting test " .. iCase)
		
	self.bRunningTests = true
	
	self:LoadSeparator(-1, "-", self.sPlayerName)
	self:LoadHealerList(-1, "Jeyp - Sashä - Kazolk - Sbobinard - Maedbh - Gtamos - Dombass - Bloodghast - Nauti - Twixh - Bretøn", self.sPlayerName)
	
	if iCase == 1 then
		self:SetPlayerCD("JEYP", 3)
		self:SetPlayerCD("MAEDBH", 13)
		self:SetPlayerCD("SBOBINARD", 6)
		self:SetPlayerCD("GTAMOS", 23)
		self:SetPlayerCD("DOMBASS", 25)
		self:SetPlayerCD("BLOODGHAST", 35)
		self:SetPlayerCD("TWIXH", 44)
		self:SetPlayerCD("NAUTI", 46)
		self:SetPlayerCD("BRETØN", 53)
		self:SetPlayerCD("SASHÄ", 59)
		
		self:ScheduleTimer(function() 
			self:SetPlayerCD("KAZOLK", 60)
		end, 4)
		self:ScheduleTimer(function() 
			self:SetPlayerCD("JEYP", 60)
		end, 5)
	elseif iCase == 2 then
		self:SetPlayerCD("NAUTI", 2)
		self:SetPlayerCD("TWIXH", 4)
		self:SetPlayerCD("SASHÄ", 16)
		self:SetPlayerCD("BRETØN", 17)
		self:SetPlayerCD("KAZOLK", 28)
		self:SetPlayerCD("JEYP", 29)
		self:SetPlayerCD("SBOBINARD", 32)
		self:SetPlayerCD("MAEDBH", 39)
		self:SetPlayerCD("GTAMOS", 49)
		self:SetPlayerCD("DOMBASS", 51)
		
		self:ScheduleTimer(function() 
			self:SetPlayerCD("BLOODGHAST", 60)
		end, 1)
		self:ScheduleTimer(function() 
			self:SetPlayerCD("TWIXH", 60)
		end, 8)
		self:ScheduleTimer(function() 
			self:SetPlayerCD("NAUTI", 60)
		end, 10)
	end
	
	self:ScheduleTimer(function() 
		self.bRunningTests = false
	end, 60)
end

function JeypLoathebHealSorter:SetPlayerCD(sPlayerUpper, iCD)
	self.tHealerList[sPlayerUpper].bCanCast = false
	self.tHealerList[sPlayerUpper].fCooldownEnd = GetServerTime() + iCD
	if sPlayerUpper == self.tHealerOrder[1] then 
		self:GoToNextPlayer()
	else
		self:UpdateFrame()
	end
end
	
function JeypLoathebHealSorter:SetSpellNames()
	local tSpellNames = nil
	if self.db.global.bTestMode then 
		tSpellNames = tTestSpells
	else
		tSpellNames = tLoathebSpells
	end
	
	JeypLoathebHealSorter.sActiveSpellName = tSpellNames.sActiveSpellName
	JeypLoathebHealSorter.sCooldownSpellName = tSpellNames.sCooldownSpellName
	JeypLoathebHealSorter.sSporeSpellName = tSpellNames.sSporeSpellName
end

function JeypLoathebHealSorter:OnEnable()
	JeypLoathebHealSorter_Frame:ClearAllPoints()
	JeypLoathebHealSorter_Frame:SetPoint("TOPLEFT", self.db.global.tFramePosition.fX, self.db.global.tFramePosition.fY)
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	self:RegisterEvent("ZONE_CHANGED_NEW_AREA")
	self:RegisterEvent("GROUP_ROSTER_UPDATE")
	self:RegisterEvent("PLAYER_REGEN_ENABLED")
	self:CheckIfActiveAlreadyPresent()
	
	self:ChecksInNaxx()	
end

function JeypLoathebHealSorter:CheckIfActiveAlreadyPresent()
	if self:GetUnitAura("player", self.sActiveSpellName) then 
		self:EnableWindow()
		self:SendHealerList()
		
		if not self.db.global.bTestMode then
			self.bSetLoadedInFight = true -- If buff is present when loading then status of cooldown can be unreliable so prevent broadcasting
		end
		
		self:SendCommMessage(self.commPrefix, "ReqHealerList", "RAID")
	else
		self:DisableWindow(true)
	end
end

function JeypLoathebHealSorter:GROUP_ROSTER_UPDATE(...) 
	if self.cRosterUpdateTimer == nil then
		self.cRosterUpdateTimer = self:ScheduleTimer(function() 
			if self.bProfiling then 
				self.fRosterUpdateTime = debugprofilestop() 
			end
			if self.iGreatestTimestamp == self.db.global.iDataTimestamp then
				self:VerifyHealerList()
			end
			
			self:SendHealerList()	
			if self.bProfiling then 
				print(string.format("[JeypLoathebHealSorter] RosterUpdate Time : %.2fms", (debugprofilestop() - self.fRosterUpdateTime)))
			end
			self.cRosterUpdateTimer = nil
		end, 0.5)
	end
end

function JeypLoathebHealSorter:PLAYER_REGEN_ENABLED(...) 
	self:AddToLogFn("PLAYER_REGEN_ENABLED") 
	JeypLoathebHealSorter:DisableWindow()
end

function JeypLoathebHealSorter:ZONE_CHANGED_NEW_AREA(...) 
	self:ChecksInNaxx()	
end

function JeypLoathebHealSorter:ChecksInNaxx()
	self.bIsInNaxx = (GetZoneText() == "Naxxramas")
	if self.bIsInNaxx then 
		if self.db.global.bTestMode then
			print("[JeypLoathebHealSorter] " .. L["Warning : Test Mode is currently enabled. Please disable it before engaging Loatheb."])
		end
		self:VerifyHealerList()
	end
end

function JeypLoathebHealSorter:COMBAT_LOG_EVENT_UNFILTERED() 
	local timestamp, event, hideCaster, sourceGUID, sourceName, sourceFlags, sourceRaidFlags, destGUID, destName, destFlags, destRaidFlags, param12, param13, param14, param15, param16, param17, param18, param19 = CombatLogGetCurrentEventInfo()
	local tCLEUParams = {
		sourceName = sourceName,
		sourceGUID = sourceGUID,
		destName = destName,
		destGUID = destGUID,
	}
	local sDestNameUpper = ""
	if destName ~= nil then
		sDestNameUpper = string.upper(destName)
	end
	
	if event == "SPELL_AURA_APPLIED" or event == "SPELL_AURA_REFRESHED" then 
		local spellId, spellName, spellSchool = param12, param13, param14
		tCLEUParams.spellName = spellName
		if destGUID == self.sPlayerGUID and self.sActiveSpellName == spellName and (self.db.global.bTestMode or destGUID ~= sourceGUID) then -- Both debuffs have the same name in English, but can be differenciated by who casts the debuff. The 2min debugg is cast (and refreshed) by Loatheb. The 1min silence is cast by the player.
			self:AddToLogCLEU(event, timestamp, tCLEUParams)
			JeypLoathebHealSorter:EnableWindow()
			if event == "SPELL_AURA_APPLIED" then 
				-- First application - Send Healer List
				self:SendHealerList()
				-- Check that everyone is alive
				if self.tHealerList then
					for sPlayer, tPlayerData in pairs(self.tHealerList) do 
						tPlayerData.bIsDead = UnitIsDeadOrGhost(sPlayer)
					end
				end
				if self.tNonHealerList then
					for sPlayer, tPlayerData in pairs(self.tNonHealerList) do 
						tPlayerData.bIsDead = UnitIsDeadOrGhost(sPlayer)
					end
				end
			end
		elseif self.sCooldownSpellName == spellName and (self.db.global.bTestMode or destGUID == sourceGUID) then 
			if self.tHealerList[sDestNameUpper] ~= nil or self.tNonHealerList[sDestNameUpper] ~= nil then
				self:AddToLogCLEU(event, timestamp, tCLEUParams)
				self:AddToLogCD(sDestNameUpper)
				
				if self.tHealerList[sDestNameUpper] ~= nil then
					self:SetPlayerCD(sDestNameUpper, 60)
					self:SendOrderUpdate()
					self:AddToLogHealerList()
				else
					self.tNonHealerList[sDestNameUpper].bCanCast = false
					self.tNonHealerList[sDestNameUpper].fCooldownEnd = GetServerTime() + 60
					self:UpdateFrame()
				end
			end
		elseif self.sSporeSpellName == spellName and self.tHealerList[sDestNameUpper] ~= nil then 
			self.tHealerList[sDestNameUpper].bHasCritBuff = true
			self.tHealerList[sDestNameUpper].fCritBuffEnd = GetServerTime() + 90
		end
	elseif event == "SPELL_AURA_REMOVED" or event == "SPELL_AURA_BROKEN"  or event == "SPELL_AURA_BROKEN_SPELL" then 
		local spellId, spellName, spellSchool = param12, param13, param14
		tCLEUParams.spellName = spellName
		if destGUID == self.sPlayerGUID and self.sActiveSpellName == spellName and self.db.global.bTestMode then 
			self:AddToLogCLEU(event, timestamp, tCLEUParams)
			JeypLoathebHealSorter:DisableWindow()
		elseif self.sCooldownSpellName == spellName and (self.db.global.bTestMode or destGUID == sourceGUID) then 
			if self.tHealerList[sDestNameUpper] ~= nil or self.tNonHealerList[sDestNameUpper] ~= nil then
				self:AddToLogCLEU(event, timestamp, tCLEUParams)
				
				if self.tHealerList[sDestNameUpper] ~= nil then
					self.tHealerList[sDestNameUpper].bCanCast = true
					self.tHealerList[sDestNameUpper].fCooldownEnd = 0
					self:AddToLogHealerList()
				else
					self.tNonHealerList[sDestNameUpper].bCanCast = true
					self.tNonHealerList[sDestNameUpper].fCooldownEnd = 0
				end
				self:UpdateFrame()
			end
		elseif self.sSporeSpellName == spellName and self.tHealerList[sDestNameUpper] ~= nil then 
			self:AddToLogCLEU(event, timestamp, tCLEUParams)
			self.tHealerList[sDestNameUpper].bHasCritBuff = false
			self.tHealerList[sDestNameUpper].fCritBuffEnd = 0
		end
	elseif event == "UNIT_DIED" then
		if self.tHealerList[sDestNameUpper] ~= nil or self.tNonHealerList[sDestNameUpper] ~= nil then
			self:AddToLogCLEU(event, timestamp, tCLEUParams)
			if self.tHealerList[sDestNameUpper] ~= nil then
				self.tHealerList[sDestNameUpper].bIsDead = true
				if self.cooldownRefreshTimer ~= nil then
					if sDestNameUpper == self.tHealerOrder[1] then 
						self:GoToNextPlayer()
					else
						self:UpdateFrame()
					end
				end
				self:AddToLogHealerList()
			else
				self.tNonHealerList[sDestNameUpper].bIsDead = true
				self:UpdateFrame()
			end
		end
	elseif event == "SPELL_RESURRECT" then
		if self.tHealerList[sDestNameUpper] ~= nil or self.tNonHealerList[sDestNameUpper] ~= nil then
			self:AddToLogCLEU(event, timestamp, tCLEUParams)
			if self.tHealerList[sDestNameUpper] ~= nil then
				self.tHealerList[sDestNameUpper].bIsDead = false
			else
				self.tNonHealerList[sDestNameUpper].bIsDead = false
			end
			self:UpdateFrame()
		end
	end
end

function JeypLoathebHealSorter:SendHealerList()
	--print("SendHealerList")
	if self:CanSendData() then 
		self:AddToLogFn("SendHealerList") 
		self:SendCommMessage(self.commPrefix, "Separator," .. self.db.global.iDataTimestamp .. "," .. self.db.global.sSeparator, "RAID")
		self:SendCommMessage(self.commPrefix, "HealerList," .. self.db.global.iDataTimestamp .. "," .. self.db.global.sHealerList, "RAID")
	end
end

function JeypLoathebHealSorter:SendOrderUpdate()
	if self:CanSendData() and self.iGreatestTimestamp == self.db.global.iDataTimestamp then 
		self:AddToLogFn("SendOrderUpdate") 
		local sOrderData = ""
		for i, sPlayer in pairs(self.tHealerOrder) do 
			sOrderData = sOrderData .. sPlayer .. ":" .. self.tHealerList[sPlayer].fCooldownEnd .. ","
		end
		sOrderData:sub(1, #sOrderData - 1)
		
		self:SendCommMessage(self.commPrefix, "OrderUpdate," .. self.db.global.iDataTimestamp .. "," .. sOrderData, "RAID")
	end
end

function JeypLoathebHealSorter:VerifyHealerList()
	if GetZoneText() == "Naxxramas" then 
		for sPlayer, tData in pairs(self.tHealerList) do 
			if not UnitInRaid(sPlayer) then
				print(string.format("[JeypLoathebHealSorter] " .. L["Warning : Player %s in healer list not found in raid."], sPlayer))
			end
		end
	end
end

function JeypLoathebHealSorter:PrepareEncounter() 
	self:AddToLogFn("PrepareEncounter") 
	self:SendHealerList()
	if self:CanSendData() then 
		self:SendCommMessage(self.commPrefix, "PrepareEncounter", "RAID")
	elseif self.db.global.bSendData then
		print("[JeypLoathebHealSorter] " .. L["Warning : Data hasn't been broadcasted because you don't have an assist."])
	end
end
	
function JeypLoathebHealSorter:OnCommReceived(prefix, message, distribution, sender)
	local tMessageData = {}
	local sSenderUpper = string.upper(sender)
	for match in string.gmatch(message, "[^,]+") do
		if tMessageData[3] ~= nil then 
			tMessageData[3] = tMessageData[3] .. "," ..match -- Separator can be a comma, so just append everything
		else
			table.insert(tMessageData, match)
		end
	end
	
	if tMessageData[1] == "Separator" and tMessageData[3] == nil then 
		-- if message is nil then we can assume that the separator is a comma and just got used as a separator for the string split
		tMessageData[3] = ","
	end
	
	self:AddToLogComm(sSenderUpper, tMessageData[1], tMessageData[2], tMessageData[3]) 
	
	--print(prefix, distribution, sSenderUpper, "======", tMessageData[1], tMessageData[2], tMessageData[3])
	if tMessageData[1] == "ReqHealerList" then 
		self:SendHealerList()
	elseif tMessageData[1] == "HealerList" then 
		self:LoadHealerList(tonumber(tMessageData[2]), tMessageData[3], sSenderUpper)
		self:SendCommMessage(self.commPrefix, "VersionID," .. self.db.global.iDataTimestamp .. "," .. self.sAddonVersionID, "RAID")
		self:AddToLogHealerList()
		--print("Received List : ", tMessageData[2], tMessageData[3]);
	elseif tMessageData[1] == "OrderUpdate" then 
		self:LoadOrderUpdate(tonumber(tMessageData[2]), tMessageData[3], sSenderUpper)
		self:AddToLogHealerList()
	elseif tMessageData[1] == "Separator" then 
		self:LoadSeparator(tonumber(tMessageData[2]), tMessageData[3], sSenderUpper)
	elseif tMessageData[1] == "PrepareEncounter" then 
		self:EnableWindow()
	elseif tMessageData[1] == "VersionID" then
		if self.tHealerList[sSenderUpper] ~= nil then 
			self.tHealerList[sSenderUpper].sAddonVersionID = tMessageData[3]
		end
	else 
		--print(prefix, distribution, sSenderUpper, "======", tMessageData[1], tMessageData[2])
	end
end

function JeypLoathebHealSorter:EnableWindow()
	if UnitInRaid("player") then 
		self:AddToLogFn("EnableWindow") 
		JeypLoathebHealSorter_Frame:Show()
		JeypLoathebHealSorter_Frame:SetAlpha(1);
	
		if self.cooldownRefreshTimer == nil then
			self.cooldownRefreshTimer = self:ScheduleRepeatingTimer(function() 
				self:UpdateTimers()		
			end, 0.5)
		end
	elseif self.db.global.bTestMode then
		print("[JeypLoathebHealSorter] " .. L["Warning : Enter Raid Mode to Enable Window."])
	end
end

function JeypLoathebHealSorter:DisableWindow(bInstant)
	if self.bAnchorVisible then 
		return
	end
	if not bInstant and self.cooldownRefreshTimer == nil then
		-- Don't play hide animation if already hidden
		return
	end

	self:AddToLogFn("DisableWindow") 
		
	if bInstant then 
		JeypLoathebHealSorter_Frame:Hide()
	else 
		UIFrameFadeOut(JeypLoathebHealSorter_Frame, 1.0, 1.0, 0.0)
	end
	if self.cooldownRefreshTimer then 
		self:CancelTimer(self.cooldownRefreshTimer)
		self.cooldownRefreshTimer = nil
	end
end

function JeypLoathebHealSorter:IconClick()
	if JeypLoathebHealSorter.Dialog.OpenFrames["JeypLoathebHealSorter"] then
		JeypLoathebHealSorter.Dialog:Close("JeypLoathebHealSorter")
	else
		JeypLoathebHealSorter.Dialog:Open("JeypLoathebHealSorter")
		JeypLoathebHealSorter.Dialog.OpenFrames["JeypLoathebHealSorter"]:EnableResize(false)
	end
end

function JeypLoathebHealSorter:DrawTooltip() 
	GameTooltip:ClearLines();
	GameTooltip:AddLine("|cff2256aeJeypLoathebHealSorter|r")
	GameTooltip:Show()
end

function JeypLoathebHealSorter:GoToNextPlayer(bSkipCallouts)
	local sNewPlayer = self:GetNextPlayer(self.tHealerOrder[1], {})
	
	self:AddToLogFn("GoToNextPlayer") 
	if sNewPlayer == self.tHealerOrder[1] then 
		-- Couldn't find any valid Player Just pick next one!
		self.tHealerOrder[1] = self.tHealerList[self.tHealerOrder[1]].sNextPlayer
	else
		self.tHealerOrder[1] = sNewPlayer
	end
	self:UpdateFrame()
	
	if bSkipCallouts ~= true then
		if self.cooldownRefreshTimer ~= nil and self.db.global.iRaidWarningCallouts ~= 0 and self:HasAssist() and self.iGreatestTimestamp == self.db.global.iDataTimestamp then
			local iFreeHealers = 0
			for i, sPlayer in pairs(self.tHealerOrder) do 
				if (not self.tHealerList[sPlayer].bIsDead and not self.tHealerList[sPlayer].bNotInRaid and self.tHealerList[sPlayer].fCooldownEnd == 0) then 
					iFreeHealers = iFreeHealers + 1
				end
			end
			if self.db.global.iRaidWarningCallouts == 1 then
				local sCurrentTime = GetServerTime()
				local sFirstHealerName = self.tHealerList[self.tHealerOrder[1]].sNameWithCase
				if self.tHealerList[self.tHealerOrder[1]].fCooldownEnd > 0 then 
					sFirstHealerName = string.format(sFirstHealerName .. L[" (CD : %is)"], self.tHealerList[self.tHealerOrder[1]].fCooldownEnd - sCurrentTime)
				end
				local sSecondHealerName = self.tHealerList[self.tHealerOrder[2]].sNameWithCase
				if self.tHealerList[self.tHealerOrder[2]].fCooldownEnd > 0 then 
					sSecondHealerName = string.format(sSecondHealerName .. L[" (CD : %is)"], self.tHealerList[self.tHealerOrder[2]].fCooldownEnd - sCurrentTime)
				end
				local sCallout = string.format(L["Current Healer : %s, Next Healer : %s, Heals available : %i"], sFirstHealerName, sSecondHealerName, iFreeHealers)
				SendChatMessage(sCallout, "RAID_WARNING")
			elseif  self.db.global.iRaidWarningCallouts == 2 then
				local sCurrentTime = GetServerTime()
				local sFirstHealerName = self.tHealerList[self.tHealerOrder[1]].sNameWithCase
				local sSecondHealerName = self.tHealerList[self.tHealerOrder[2]].sNameWithCase
				if sFirstHealerName ~= nil then
					SendChatMessage(L["Your turn to heal !"], "WHISPER", nil, sFirstHealerName)
				end
				if sSecondHealerName ~= nil then
					SendChatMessage(L["Get ready to heal !"], "WHISPER", nil, sSecondHealerName)
				end
			end
		end
	end
end

function JeypLoathebHealSorter:GetHealerOrderString()
	local sHealerOrder = ""
	local fTime = GetServerTime()
	for i, sPlayer in pairs(self.tHealerOrder) do 
		sHealerOrder = sHealerOrder .. sPlayer
		if self.tHealerList[sPlayer].fCooldownEnd > 0 then 
			sHealerOrder = sHealerOrder .. "(" .. math.floor(self.tHealerList[sPlayer].fCooldownEnd - fTime) .. "s CD)"
		end
		sHealerOrder = sHealerOrder .. ", "
	end
	return sHealerOrder:sub(1, #sHealerOrder - 2)
end

function JeypLoathebHealSorter:GetNextPlayer(sPlayer, tFilteredPlayers)
	if sPlayer == nil or self.tHealerList[sPlayer] == nil then 
		return nil
	end
	
	local sDebugData = ""
	local fTime = GetServerTime()
	
	if JeypLoathebHealSorter.bPrintDebugInfo then
		sDebugData = "=====================================|nHealer Order : ".. self:GetHealerOrderString() .."|n["..sPlayer.."] : "
	end
	
	local i = 1
	local sNextPlayer = self.tHealerList[sPlayer].sNextPlayer
	local sLowestPlayer = nil
	local bLowestIsNotInRaid = false
	local bLowestIsCD = false
	local fLowestCD = fTime + 65
	while (not self.tHealerList[sNextPlayer].bCanCast) or self.tHealerList[sNextPlayer].bIsDead or self.tHealerList[sNextPlayer].bNotInRaid or tFilteredPlayers[sNextPlayer] do 
		sDebugData = sDebugData .. "|n" .. sNextPlayer .. "?"
		if not tFilteredPlayers[sNextPlayer] then 
			if sLowestPlayer == nil then
				sLowestPlayer = sNextPlayer
				sDebugData = sDebugData .. " NIL"
				if self.tHealerList[sNextPlayer].bNotInRaid then 
					bLowestIsNotInRaid = true
					sDebugData = sDebugData .. " NIL-NotInRaid"
				elseif self.tHealerList[sNextPlayer].fCooldownEnd > 0 then 
					bLowestIsCD = true
					fLowestCD = self.tHealerList[sNextPlayer].fCooldownEnd
					sDebugData = sDebugData .. " NIL-CD" .. (self.tHealerList[sNextPlayer].fCooldownEnd - fTime)
				end
			elseif not (bLowestIsCD or bLowestIsNotInRaid) and self.tHealerList[sNextPlayer].bIsDead then 
				sLowestPlayer = sNextPlayer
				sDebugData = sDebugData .. " Dead"
			elseif not bLowestIsCD and self.tHealerList[sNextPlayer].bNotInRaid then
				if not bLowestIsNotInRaid then
					bLowestIsNotInRaid = true
					sLowestPlayer = sNextPlayer
					sDebugData = sDebugData .. " NotInRaid"
				end
			elseif (self.tHealerList[sNextPlayer].fCooldownEnd > 0 and (fLowestCD > self.tHealerList[sNextPlayer].fCooldownEnd)) then 
				bLowestIsCD = true
				sDebugData = sDebugData .. " CD" .. (self.tHealerList[sNextPlayer].fCooldownEnd - fTime) .. " - previous lowest is " .. (fLowestCD - fTime)
				fLowestCD = self.tHealerList[sNextPlayer].fCooldownEnd
				sLowestPlayer = sNextPlayer
			else
				sDebugData = sDebugData .. " Nothing(" .. (self.tHealerList[sNextPlayer].fCooldownEnd - fTime) .. " of CD - lowest is " .. (fLowestCD - fTime) .. ")"
			end 
		else
			sDebugData = sDebugData .. " Filtered"
		end

		if self.tHealerList[sNextPlayer].sNextPlayer == sPlayer then
			if JeypLoathebHealSorter.bPrintDebugInfo then
				local sLowestPlayerName = sLowestPlayer and sLowestPlayer or "nil"
				print(sDebugData .. "|n Return ".. sLowestPlayerName)
			end
			return sLowestPlayer 
		end
		
		i = i + 1
		sNextPlayer = self.tHealerList[sNextPlayer].sNextPlayer
	end
	
	if JeypLoathebHealSorter.bPrintDebugInfo then
		print(sDebugData .. "|n Finish ".. sNextPlayer)
	end
	return sNextPlayer
end

function JeypLoathebHealSorter:PlayerNameWithClassColor(sName) 
	if sName == nil then return "" end
	
	local sClass = select(2, UnitClass(sName))
	if sClass == nil then 
		return sName
	else 
		return "|c" .. select(4, GetClassColor(sClass)) .. sName .. "|r"
	end
end

function JeypLoathebHealSorter:SPairs(t) 
    local keys = {}
    for k in pairs(t) do keys[#keys+1] = k end
    
    table.sort(keys)
    local i = 0
    return function()
        i = i + 1
        if keys[i] then
            return keys[i], t[keys[i]]
        end
    end
end

function JeypLoathebHealSorter:SplitString(sString, sSeparator)
	local tSubStrings = {}
	for match in string.gmatch(sString, "[^" .. sSeparator .. "]+") do
	  table.insert(tSubStrings, match)
	end
	return tSubStrings
end

function JeypLoathebHealSorter:GetUnitAura(unit, spell, filter)
  for i = 1, 255 do
    local name, _, _, _, _, _, _, _, _, spellId = UnitAura(unit, i, filter)
    if name == nil then return end
    if spell == spellId or spell == name then
      return UnitAura(unit, i, filter)
    end
  end
end